// ex: set tabstop=2 shiftwidth=2 expandtab syn=c:
// $Id: internal_interrupt.regular.c,v 1.4 2006/03/09 02:31:43 wangf Exp $

/*                                                                      
 * "Copyright (c) 2000-2003 The Regents of the University  of California.  
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice, the following
 * two paragraphs and the author appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Copyright (c) 2002-2003 Intel Corporation
 * All rights reserved.
 *
 * This file is distributed under the terms in the attached INTEL-LICENSE     
 * file. If you do not find these files, copies can be found by writing to
 * Intel Research Berkeley, 2150 Shattuck Avenue, Suite 1300, Berkeley, CA, 
 * 94704.  Attention:  Intel License Inquiry.
 */

/*
 * Authors:  Rodrigo Fonseca
 * Date Last Modified: 2005/05/26
 */

/* Allows events with global knowledge independent of the actual code
   on the motes. This is a mechanism that allows powerful scripting,
   by inserting identified events in the event queue.*/


#include <S4.h>
#include <S4Command.h>

/* Description: */
/* This simulation generates route request to random pairs of nodes at a constant rate */
/* After 5 hours, we start killing one mote 15 minutes */

enum {
  INT_EVENT_FIRST = 0x8000,
  INT_EVENT_SEND_MSG,
  INT_EVENT_KILL_MOTE,
};

/* Internal Functions */

enum {
  MSGS_INITIAL_WAIT = 600,  //wait for 10 minutes to stabilize
  MSGS_INTERVAL = 1,         //send one message per second
  KILL_INITIAL_WAIT = 15000,  //30 minutes after the routes start
  KILL_INTERVAL = 600,       //kill one mote every 10 minutes
};

//Forward Declarations
#ifdef S4ROUTING
result_t select_random_routing_pair(short *from, short *to, 
     Coordinates ** coords_from, Coordinates ** coords_to);
#endif
#ifdef CRROUTING
result_t select_random_routing_pair(short *from, short *to, uint8_t *beacon);
#endif
long int get_next_route_interval();
#ifdef S4ROUTING
void send_route_to_command(short origin, short dest, Coordinates * coords_to);
#endif
#ifdef CRROUTING
void send_route_to_command(short origin, short dst, uint8_t beacon);
#endif
result_t select_mote_to_kill(short *mote);
long long get_next_kill_interval();



void schedule_first_events() {
  //dbg_clear(DBG_ROUTE,"II:schedule_first_events, at %llu\n",tos_state.tos_time);
  scheduleInterrupt(INT_EVENT_SEND_MSG, (long long)MSGS_INITIAL_WAIT*4000000);
  scheduleInterrupt(INT_EVENT_KILL_MOTE, (long long)KILL_INITIAL_WAIT*4000000);
}

void send_route_to_random_pair() {
  short from,to;
#ifdef S4ROUTING
  Coordinates *coords_from,*coords_to;
#endif
#ifdef CRROUTING
  uint8_t beacon;
#endif
  //dbg_clear(DBG_ROUTE,"II:Choosing a random pair to route\n");
#ifdef S4ROUTING
  if (select_random_routing_pair(&from,&to,&coords_from,&coords_to) == SUCCESS) {
#endif
#ifdef CRROUTING
  if (select_random_routing_pair(&from,&to,&beacon) == SUCCESS) {
#endif
    dbg_clear(DBG_ROUTE,"II:Selected pair: from %d to %d (closest beacon %d)\n",from,to,beacon+1);
#ifdef S4ROUTING
    dbg_clear(DBG_USR3,"II:Coordinates for dest mote %d: ",to);
    coordinates_print(DBG_USR3,coords_to);
    //Send route command
    send_route_to_command(from,to,coords_to);
#endif
#ifdef CRROUTING
    send_route_to_command(from,to,beacon);
#endif
  } else {
    dbg_clear(DBG_ROUTE,"II:Failed to select pair, skipping slot\n");
  }
  scheduleInterrupt(INT_EVENT_SEND_MSG, tos_state.tos_time + get_next_route_interval());
}

/* Every one second...*/
long int get_next_route_interval() {
  return (long int)4000000*MSGS_INTERVAL;
}

//choose a mote that is alive and is not a beacon
void periodic_kill_motes() {
  short victim;
  if (select_mote_to_kill(&victim)) {
   dbg_clear(DBG_ROUTE,"II: Killing mote %d at %llu\n",victim,tos_state.tos_time);
    nido_stop_mote(victim); 
    scheduleInterrupt(INT_EVENT_KILL_MOTE, tos_state.tos_time + get_next_kill_interval());
  }
  //else stop: no one else to kill
}

long long get_next_kill_interval() {
  return (long long)4000000*KILL_INTERVAL;
}

#ifdef S4ROUTING
result_t select_random_routing_pair(short *from, short *to,
                          Coordinates ** coords_from, Coordinates ** coords_to) {
#endif
#ifdef CRROUTING
result_t select_random_routing_pair(short *from, short *to, uint8_t *beacon) {
#endif
  //select random dest
  int i,c,d;
  bool found;
  char *name = "S4StateM$my_coords";
  uintptr_t p;
  size_t size;
#ifdef CRROUTING
  Coordinates * coords_to;
  Coordinates * coords_from;
#endif
  
  d = (int)(((double)tos_state.num_nodes)*rand()/(RAND_MAX+1.0));
  //check that mote is on and has valid coordinates of at least mask
  found = 0;
  for (i = 0; i < tos_state.num_nodes && !found; i++) {
    c = (d + i) % tos_state.num_nodes;
    //get coordinates for this mote and count valid components 
    if (__nesc_nido_resolve(c, name, &p, &size) != 0) {
      dbg(DBG_ROUTE,"II: Error resolving %s for mote %d\n",name,c);
      continue;
    }
#ifdef S4ROUTING
    *coords_to = (Coordinates*)p;
    found = (tos_state.moteOn[c] &&
             coordinates_count_valid_components(*coords_to) == N_ROOT_BEACONS);
    if (found)
      *to = c;
#endif
#ifdef CRROUTING
    coords_to = (Coordinates*)p;
    found = (tos_state.moteOn[c] && 
             coordinates_count_valid_components(coords_to) == N_ROOT_BEACONS);
    if (found) {
      *to = c;
      *beacon = coordinates_get_closest_beacon(coords_to);
    }
#endif
  }
  if (!found)
    return FAIL;
  //look for a source
  d = (int)(((double)tos_state.num_nodes)*rand()/(RAND_MAX+1.0));
  //check that mote is on and has valid coordinates of at least mask

  found = 0;
  for (i = 0; i < tos_state.num_nodes && !found; i++) {
    c = (d + i) % tos_state.num_nodes;
    if (c == *to) continue;
    //get coordinates for this mote and count valid components 
    if (__nesc_nido_resolve(c, name, &p, &size) != 0) {
      dbg(DBG_ROUTE,"II: Error resolving %s for mote %d\n",name,c);
      continue;
    }
#ifdef S4ROUTING
    *coords_from = (Coordinates*)p;
    found = (tos_state.moteOn[c] &&
             coordinates_count_valid_components(*coords_from) == N_ROOT_BEACONS);
#endif
#ifdef CRROUTING
    coords_from = (Coordinates*)p;
    found = (tos_state.moteOn[c] && 
             coordinates_count_valid_components(coords_from) == N_ROOT_BEACONS);
#endif
 
    if (found) 
      *from = c;
  }
  if (!found)
    return FAIL;
  return SUCCESS;
}

#ifdef S4ROUTING
void send_route_to_command(short from, short to, Coordinates *coords) {
#endif
#ifdef CRROUTING
void send_route_to_command(short from, short to, uint8_t beacon) {
#endif
  TOS_Msg message;
  S4CommandMsg *cm = (S4CommandMsg*) &message.data;
  
  S4CommandArgs* cmd_args = 
       (S4CommandArgs*) &cm->type_data.data;

  //added by Feng Wang on Sept. 29, to increment cmd seqno
  static uint8_t i = 0;

  message.addr = from;
  message.type = AM_S4_COMMAND_MSG;
  message.group = TOS_AM_GROUP; 
  message.length = sizeof(S4CommandMsg);
  
  cm->header.last_hop = TOS_UART_ADDR;
  cm->header.seqno = 0;
  
  cm->type_data.hopcount = 1;
  cm->type_data.origin = TOS_UART_ADDR;
  cm->type_data.type = S4_CMD_APP_ROUTE_TO;

  cmd_args->seqno = ++i;
  cmd_args->flags = 0;

  //Destination
  cmd_args->args.dest.addr = to;
  //cmd_args->args.dest.mode = 2;
#ifdef S4ROUTING
  coordinates_copy(coords, &cmd_args->args.dest.coords);
#endif
#ifdef CRROUTING
  cmd_args->args.dest.closest_beacon = beacon;
#endif

  //Done setting up message, call 
  sendUARTMessage(from, &message);
  
}

result_t select_mote_to_kill(short *victim) {
  int d,c,i;
  bool found;

  d = (int)(((double)tos_state.num_nodes)*rand()/(RAND_MAX+1.0));
  found = 0;
  //select a mote that is alive and is not a beacon
  for (i = 0; i < tos_state.num_nodes && !found; i++) {
    c = (d + i) % tos_state.num_nodes;
    found = ( tos_state.moteOn[c] && 
             (hc_root_beacon_id[c] == INVALID_BEACON_ID)
            ); 
  }
  if (found) 
    *victim = c;
  return (found)?SUCCESS:FAIL;
}

/* --------------------------------------------------------------------*/
/* Exported Functions */

void internalInterruptInit() {
  dbg_clear(DBG_ROUTE,"II:InternalInterruptInit! %llu\n",tos_state.tos_time);
  scheduleInterrupt(INT_EVENT_FIRST, tos_state.tos_time);
}


/*This function is called whenever an interrupt event is
 *triggered. This allows us to dispatch different handlers
 *based on id. Rescheduling can be done by calling
 *scheduleInterrupt(id,time)
 */
void internalInterruptHandler(uint32_t id) {
  dbg_clear(DBG_ROUTE,"II:InternalInterruptHandler %llu!\n",tos_state.tos_time);
  switch(id) {
    case INT_EVENT_FIRST:
      schedule_first_events();
      break;
    case INT_EVENT_SEND_MSG:
      send_route_to_random_pair();
      break;
    case INT_EVENT_KILL_MOTE:
      periodic_kill_motes();
      break;
    default:
      break;
  }
}


